// Copyright Inspirel Sp. z o.o. 2014.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h>

#include "YAMI4.h"

/* max length of header type tag                  */
/* (normally you do not want to change this,      */
/* this is part of the high-level YAMI4 protocol) */
#define MAX_HEADER_TYPE_LENGTH 13u

/* max length of object name */
#define MAX_OBJECT_NAME_LENGTH 20u

/* max length of message name */
#define MAX_MESSAGE_NAME_LENGTH 20u

/* max length of exception (rejection) message */
#define MAX_EXCEPTION_REASON_LENGTH 100u

#define FRAME_HEADER_SIZE 16u

static size_t yami_round_up_to_4(const size_t v)
{
    size_t res = v;
    const size_t rem = v % 4u;
    switch (rem)
    {
    case 1u:
        res += 3u;
        break;
    case 2u:
        res += 2u;
        break;
    case 3u:
        res += 1u;
        break;
    default:
        /* already at the 4 boundary */
        ;
    }
    
    return res;
}

static size_t yami_round_up_to_8(const size_t v)
{
    size_t res = v;
    const size_t rem = v % 8u;
    switch (rem)
    {
    case 1u:
        res += 7u;
        break;
    case 2u:
        res += 6u;
        break;
    case 3u:
        res += 5u;
        break;
    case 4u:
        res += 4u;
        break;
    case 5u:
        res += 3u;
        break;
    case 6u:
        res += 2u;
        break;
    case 7u:
        res += 1u;
        break;
    default:
        /* already at the 8 boundary */
        ;
    }
    
    return res;
}

static size_t yami_strlen(const char s[])
{
    size_t i = 0u;
    
    while ((int32_t)s[i] != (int32_t)'\0')
    {
        ++i;
    }
    
    return i;
}

static int32_t yami_strncmp(const char s1[], const char s2[], const size_t n)
{
    int32_t res = 0;
    size_t i = 0u;
    
    while (i < n)
    {
        const int32_t c1 = (int32_t)s1[i];
        const int32_t c2 = (int32_t)s2[i];
        
        if (c1 != c2)
        {
            res = c1 - c2;
            break;
        }
        
        ++i;
    }
    
    return res;
}

static void yami_memcpy(uint8_t dest[], const size_t dest_from,
    const uint8_t src[], const size_t src_from, const size_t n)
{
    size_t i;
    
    for (i = 0u; i != n; ++i)
    {
        dest[dest_from + i] = src[src_from + i];
    }
}

enum yami_result yami_put_type(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const enum yami_parameter_type type)
{
    enum yami_result res;
    
    switch (type)
    {
    case yami_boolean:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 1);
        break;
    case yami_integer:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 2);
        break;
    case yami_long_long:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 3);
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 4);
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 5);
        break;
    case yami_binary:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 6);
        break;
    case yami_boolean_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 7);
        break;
    case yami_integer_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 8);
        break;
    case yami_long_long_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 9);
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 10);
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 11);
        break;
    case yami_binary_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 12);
        break;
    case yami_nested_parameters:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 13);
        break;
    case yami_nested_parameters_array:
        res = yami_put_integer(buffer, buffer_size, buffer_position, 14);
        break;
    default:
        res = yami_unexpected_value;
    }
    
    return res;
}

enum yami_result yami_get_type(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    enum yami_parameter_type * const type)
{
    enum yami_result res;
    int32_t code;
    
    res = yami_get_integer(buffer, buffer_size, buffer_position, &code);
    if (res == yami_ok)
    {
        switch (code)
        {
        case 1:
            *type = yami_boolean;
            break;
        case 2:
            *type = yami_integer;
            break;
        case 3:
            *type = yami_long_long;
            break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
        case 4:
            *type = yami_double_float;
            break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        case 5:
            *type = yami_string;
            break;
        case 6:
            *type = yami_binary;
            break;
        case 7:
            *type = yami_boolean_array;
            break;
        case 8:
            *type = yami_integer_array;
            break;
        case 9:
            *type = yami_long_long_array;
            break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
        case 10:
            *type = yami_double_float_array;
            break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        case 11:
            *type = yami_string_array;
            break;
        case 12:
            *type = yami_binary_array;
            break;
        case 13:
            *type = yami_nested_parameters;
            break;
        case 14:
            *type = yami_nested_parameters_array;
            break;
        default:
            res = yami_unexpected_value;
        }
    }
    
    return res;
}

enum yami_result yami_put_integer(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 4u) /* sizeof(value) == 4 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        uint8_t byte0;
        uint8_t byte1;
        uint8_t byte2;
        uint8_t byte3;
        
        uint32_t uvalue;
        if (value >= 0)
        {
            uvalue = (uint32_t)value;
        }
        else
        {
            const int32_t tmp = -1 * (value + 1);
            uvalue = (uint32_t)tmp ^ 0xffffffffu;
        }
        
        byte0 = (uint8_t)(uvalue & 0xffu);
        byte1 = (uint8_t)((uvalue & 0xff00u) >> 8u);
        byte2 = (uint8_t)((uvalue & 0xff0000u) >> 16u);
        byte3 = (uint8_t)((uvalue & 0xff000000u) >> 24u);
        
        buffer[*buffer_position] = byte0;
        buffer[*buffer_position + 1u] = byte1;
        buffer[*buffer_position + 2u] = byte2;
        buffer[*buffer_position + 3u] = byte3;

        *buffer_position += 4u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_integer(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 4u) /* sizeof(*value) == 4 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        const uint8_t byte0 = buffer[*buffer_position];
        const uint8_t byte1 = buffer[*buffer_position + 1u];
        const uint8_t byte2 = buffer[*buffer_position + 2u];
        const uint8_t byte3 = buffer[*buffer_position + 3u];
    
        const uint32_t uvalue =
            ((uint32_t)byte3 << 24) |
            ((uint32_t)byte2 << 16) |
            ((uint32_t)byte1 << 8) |
            (uint32_t)byte0;
        
        if ((uvalue & 0x80000000u) == 0u)
        {
            /* positive number */
            
            *value = (int32_t)uvalue;
        }
        else
        {
            /* negative number */
            
            const uint32_t temp = uvalue ^ 0xffffffffu;
            *value = (int32_t)temp - 1;
        }
        
        *buffer_position += 4u;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_long_long(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(value) == 8 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        uint8_t byte0;
        uint8_t byte1;
        uint8_t byte2;
        uint8_t byte3;
        uint8_t byte4;
        uint8_t byte5;
        uint8_t byte6;
        uint8_t byte7;
        
        uint64_t uvalue;
        if (value >= 0LL)
        {
            uvalue = (uint64_t)value;
        }
        else
        {
            const int64_t tmp = -1LL * (value + 1LL);
            uvalue = (uint64_t)tmp ^ 0xffffffffffffffffull;
        }
        
        byte0 = (uint8_t)(uvalue & 0xffull);
        byte1 = (uint8_t)((uvalue & 0xff00ull) >> 8u);
        byte2 = (uint8_t)((uvalue & 0xff0000ull) >> 16u);
        byte3 = (uint8_t)((uvalue & 0xff000000ull) >> 24u);
        byte4 = (uint8_t)((uvalue & 0xff00000000ull) >> 32u);
        byte5 = (uint8_t)((uvalue & 0xff0000000000ull) >> 40u);
        byte6 = (uint8_t)((uvalue & 0xff000000000000ull) >> 48u);
        byte7 = (uint8_t)((uvalue & 0xff00000000000000ull) >> 56u);
        
        buffer[*buffer_position] = byte0;
        buffer[*buffer_position + 1u] = byte1;
        buffer[*buffer_position + 2u] = byte2;
        buffer[*buffer_position + 3u] = byte3;
        buffer[*buffer_position + 4u] = byte4;
        buffer[*buffer_position + 5u] = byte5;
        buffer[*buffer_position + 6u] = byte6;
        buffer[*buffer_position + 7u] = byte7;

        *buffer_position += 8u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_long_long(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int64_t * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(*value) == 8 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        const uint8_t byte0 = buffer[*buffer_position];
        const uint8_t byte1 = buffer[*buffer_position + 1u];
        const uint8_t byte2 = buffer[*buffer_position + 2u];
        const uint8_t byte3 = buffer[*buffer_position + 3u];
        const uint8_t byte4 = buffer[*buffer_position + 4u];
        const uint8_t byte5 = buffer[*buffer_position + 5u];
        const uint8_t byte6 = buffer[*buffer_position + 6u];
        const uint8_t byte7 = buffer[*buffer_position + 7u];

        const uint64_t uvalue =
            ((uint64_t)byte7 << 56) |
            ((uint64_t)byte6 << 48) |
            ((uint64_t)byte5 << 40) |
            ((uint64_t)byte4 << 32) |
            ((uint64_t)byte3 << 24) |
            ((uint64_t)byte2 << 16) |
            ((uint64_t)byte1 << 8) |
            (uint64_t)byte0;
        
        if ((uvalue & 0x8000000000000000ull) == 0ull)
        {
            /* positive number */
            
            *value = (int64_t)uvalue;
        }
        else
        {
            /* negative number */
            
            const uint64_t temp = uvalue ^ 0xffffffffffffffffull;
            *value = (int64_t)temp - 1LL;
        }
        
        *buffer_position += 8u;

        res = yami_ok;
    }
    
    return res;
}

#ifdef YAMI4_WITH_DOUBLE_FLOAT

enum yami_result yami_put_double_float(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const double value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(value) == 8 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        const uint8_t * p = (const uint8_t *)&value;
        
        buffer[*buffer_position] = *p++;
        buffer[*buffer_position + 1u] = *p++;
        buffer[*buffer_position + 2u] = *p++;
        buffer[*buffer_position + 3u] = *p++;
        buffer[*buffer_position + 4u] = *p++;
        buffer[*buffer_position + 5u] = *p++;
        buffer[*buffer_position + 6u] = *p++;
        buffer[*buffer_position + 7u] = *p;

        *buffer_position += 8u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_double_float(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    double * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(*value) == 8 */
    {
        res = yami_not_enough_space;
    }
    else
    {
        uint8_t * p = (uint8_t *)value;
        
        *p++ = buffer[*buffer_position];
        *p++ = buffer[*buffer_position + 1u];
        *p++ = buffer[*buffer_position + 2u];
        *p++ = buffer[*buffer_position + 3u];
        *p++ = buffer[*buffer_position + 4u];
        *p++ = buffer[*buffer_position + 5u];
        *p++ = buffer[*buffer_position + 6u];
        *p   = buffer[*buffer_position + 7u];
        
        *buffer_position += 8u;

        res = yami_ok;
    }
    
    return res;
}

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

enum yami_result yami_put_cstring(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char value[])
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* strings are prefixed with length */

    const size_t value_length = yami_strlen(value);
    
    /* length prefix with string length */
    const size_t space_needed = yami_round_up_to_4(4u + value_length);
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; i != value_length; ++i)
            {
                buffer[*buffer_position + i] = (uint8_t)value[i];
            }
            
            /* add padding to align at the word boundary if necessary */
            
            while ((i + 4u) < space_needed)
            {
                buffer[*buffer_position + i] = (uint8_t)0u;
                ++i;
            }

            *buffer_position += i;
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_cstring(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    char value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;

    if (value_length > space_left)
    {
        res = yami_not_enough_space;
    }
    else
    {
        size_t i;
        for (i = 0u; i != value_length; ++i)
        {
            value[i] = (char)buffer[*buffer_position + i];
        }
        
        value[i] = (char)'\0';
        
        *buffer_position += yami_round_up_to_4(value_length);
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_binary(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const uint8_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* binaries are prefixed with length */

    /* length prefix with binary length */
    const size_t space_needed = yami_round_up_to_4(4u + value_length);
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; i != value_length; ++i)
            {
                buffer[*buffer_position + i] = value[i];
            }
            
            /* add padding to align at the word boundary if necessary */
            
            while ((i + 4u) < space_needed)
            {
                buffer[*buffer_position + i] = (uint8_t)0u;
                ++i;
            }

            *buffer_position += i;
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_binary(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    uint8_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;

    if (value_length > space_left)
    {
        res = yami_not_enough_space;
    }
    else
    {
        size_t i;
        for (i = 0u; i != value_length; ++i)
        {
            value[i] = buffer[*buffer_position + i];
        }
        
        *buffer_position += yami_round_up_to_4(value_length);
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_boolean_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* boolean arrays are packed and are prefixed with length */

    const size_t length_padded = yami_round_up_to_8(value_length);
    const size_t space_needed = yami_round_up_to_4(4u + (length_padded / 8u));
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i = 0u; /* buffer byte index */
            size_t j = 0u; /* buffer bit index  */
            size_t k = 0u; /* value array index */
            
            buffer[*buffer_position] = (uint8_t)0u;
            while (k < value_length)
            {
                if (value[k] != 0)
                {
                    const uint32_t mask = (uint32_t)1u << j;
                    const uint32_t new_value = (uint32_t)buffer[*buffer_position + i] | mask;
                    buffer[*buffer_position + i] = (uint8_t)new_value;
                }

                ++k;
                ++j;
                if (j == 8u)
                {
                    j = 0u;
                    ++i;
                    if (k < value_length)
                    {
                        buffer[*buffer_position + i] = (uint8_t)0u;
                    }
                }
            }

            if (j != 0u)
            {
                /* the packed array finished in the middle of last byte */
                ++i;
            }
            
            while ((i + 4u) < space_needed)
            {
                buffer[*buffer_position + i] = (uint8_t)0u;
                ++i;
            }

            *buffer_position += space_needed;
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_boolean_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* boolean arrays are packed */

    const size_t length_padded = yami_round_up_to_8(value_length);
    const size_t space_needed = yami_round_up_to_4(4u + (length_padded / 8u));
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        size_t i = 0u; /* buffer byte index */
        size_t j = 0u; /* buffer bit index  */
        size_t k = 0u; /* value array index */
        
        while (k < value_length)
        {
            const uint32_t mask = (uint32_t)1u << j;
            const uint32_t b = (uint32_t)buffer[*buffer_position + i];
            
            if ((b & mask) != 0u)
            {
                value[k] = 1;
            }
            else
            {
                value[k] = 0;
            }

            ++k;
            ++j;
            if (j == 8u)
            {
                j = 0u;
                ++i;
            }
        }
        
        *buffer_position += space_needed;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_integer_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + (4u * value_length);
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
            {
                res = yami_put_integer(buffer, buffer_size, buffer_position, value[i]);
            }
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_integer_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 4u * value_length;
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        size_t i;
        res = yami_ok;
        for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
        {
            res = yami_get_integer(buffer, buffer_size, buffer_position, &value[i]);
        }
    }
    
    return res;
}

enum yami_result yami_put_long_long_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + (8u * value_length);
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
            {
                res = yami_put_long_long(buffer, buffer_size, buffer_position, value[i]);
            }
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_long_long_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int64_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 8u * value_length;
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        size_t i;
        res = yami_ok;
        for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
        {
            res = yami_get_long_long(buffer, buffer_size, buffer_position, &value[i]);
        }
    }
    
    return res;
}

#ifdef YAMI4_WITH_DOUBLE_FLOAT

enum yami_result yami_put_double_float_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const double value[], size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + 8u * value_length;
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        const int32_t len = (int32_t)value_length;
        
        res = yami_put_integer(buffer, buffer_size, buffer_position, len);
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
            {
                res = yami_put_double_float(buffer, buffer_size, buffer_position, value[i]);
            }
        }
    }
    
    return res;
}

enum yami_result yami_get_raw_double_float_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value[], size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 8u * value_length;
    
    if (space_left < space_needed)
    {
        res = yami_not_enough_space;
    }
    else
    {
        res = yami_ok;
        if (res == yami_ok)
        {
            size_t i;
            for (i = 0u; (res == yami_ok) && (i != value_length); ++i)
            {
                res = yami_get_double_float(buffer, buffer_size, buffer_position, &value[i]);
            }
        }
    }
    
    return res;
}

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

enum yami_result yami_skip_field(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const enum yami_parameter_type type)
{
    enum yami_result res;
    
    switch (type)
    {
    case yami_boolean:
    case yami_integer:
        *buffer_position += 4u;
        res = yami_ok;
        break;
    case yami_long_long:
        *buffer_position += 8u;
        res = yami_ok;
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float:
        *buffer_position += 8u;
        res = yami_ok;
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string:
    case yami_binary:
        {
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                *buffer_position += yami_round_up_to_4((size_t)len);
            }
        }
        break;
    case yami_boolean_array:
        {
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                const size_t length_padded = yami_round_up_to_8((size_t)len);
                const size_t space_taken = yami_round_up_to_4(length_padded / 8u);

                *buffer_position += space_taken;
            }
        }
        break;
    case yami_integer_array:
        {
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                *buffer_position += 4u * (size_t)len;
            }
        }
        break;
    case yami_long_long_array:
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float_array:
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        {
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                *buffer_position += 8u * (size_t)len;
            }
        }
        break;
    case yami_string_array:
    case yami_binary_array:
        {
            int32_t array_len;
            size_t i;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &array_len);
            for (i = 0u; (res == yami_ok) && (i != (size_t)array_len); ++i)
            {
                int32_t string_len;
                res = yami_get_integer(buffer, buffer_size, buffer_position, &string_len);
                if (res == yami_ok)
                {
                    *buffer_position += yami_round_up_to_4((size_t)string_len);
                }
            }
        }
        break;
    case yami_nested_parameters:
    case yami_nested_parameters_array:
    default:
        res = yami_unexpected_value;
    }
    
    return res;
}

static enum yami_result yami_fill_message_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char object_name[], const char message_name[], const int64_t message_id)
{
    enum yami_result res;

    /* message header contains 4 fields: */
    /* - message type                    */
    /* - object name                     */
    /* - message name                    */
    /* - message id                      */
    
    res = yami_put_integer(buffer, buffer_size, buffer_position, 4);
    
    if (res == yami_ok)
    {
        const char type_tag[] = "type";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            const char message_type[] = "message";
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_type);
        }
    }

    if (res == yami_ok)
    {
        const char object_name_tag[] = "object_name";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, object_name_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            res = yami_put_cstring(buffer, buffer_size, buffer_position, object_name);
        }
    }
    
    if (res == yami_ok)
    {
        const char msg_name_tag[] = "message_name";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_name_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_name);
        }
    }

    if (res == yami_ok)
    {
        const char msg_id_tag[] = "message_id";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);
        }

        if (res == yami_ok)
        {
            res = yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
        }
    }

    return res;
}

static enum yami_result yami_fill_reply_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t message_id)
{
    enum yami_result res;

    /* reply header contains 2 fields: */
    /* - message type                  */
    /* - message id                    */
    
    res = yami_put_integer(buffer, buffer_size, buffer_position, 2);
    
    if (res == yami_ok)
    {
        const char type_tag[] = "type";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            const char message_type[] = "reply";
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_type);
        }
    }

    if (res == yami_ok)
    {
        const char msg_id_tag[] = "message_id";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);
        }

        if (res == yami_ok)
        {
            res = yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
        }
    }

    return res;
}

static enum yami_result yami_fill_exception_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char reason[], const int64_t message_id)
{
    enum yami_result res;

    /* exception header contains 3 fields: */
    /* - message type                      */
    /* - message id                        */
    /* - exception reason                  */
    
    res = yami_put_integer(buffer, buffer_size, buffer_position, 3);
    
    if (res == yami_ok)
    {
        const char type_tag[] = "type";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            const char message_type[] = "exception";
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_type);
        }
    }

    if (res == yami_ok)
    {
        const char msg_id_tag[] = "message_id";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);
        }

        if (res == yami_ok)
        {
            res = yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
        }
    }

    if (res == yami_ok)
    {
        const char reason_tag[] = "reason";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, reason_tag);

        if (res == yami_ok)
        {
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            res = yami_put_cstring(buffer, buffer_size, buffer_position, reason);
        }
    }

    return res;
}

static enum yami_result yami_parse_message_header(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    char out_header_type[], const size_t header_type_size,
    char out_object_name[], const size_t object_name_size,
    char out_message_name[], const size_t message_name_size,
    char out_reason[], const size_t reason_size,
    int64_t * const out_message_id)
{
    enum yami_result res;
    int32_t tmp;
    
    const char type_tag[] = "type";
    const size_t type_tag_len = sizeof(type_tag);
    const char object_name_tag[] = "object_name";
    const size_t object_name_tag_len = sizeof(object_name_tag);
    const char message_name_tag[] = "message_name";
    const size_t message_name_tag_len = sizeof(message_name_tag);
    const char message_id_tag[] = "message_id";
    const size_t message_id_tag_len = sizeof(message_id_tag);
    const char reason_tag[] = "reason";
    const size_t reason_tag_len = sizeof(reason_tag);
    
    /* set the output parameters to "safe" values */

    out_header_type[0] = (char)'\0';
    out_object_name[0] = (char)'\0';
    out_message_name[0] = (char)'\0';
    out_reason[0] = (char)'\0';
    *out_message_id = 0LL;
    
    res = yami_get_integer(buffer, buffer_size, buffer_position, &tmp);
    if (res == yami_ok)
    {
        if (tmp > 0)
        {
            const size_t num_of_entries = (size_t)tmp;
            size_t i;
            int32_t len;
            char name[MAX_HEADER_TYPE_LENGTH]; /* 13, the length of longest expected field name */
            enum yami_parameter_type type;
            
            for (i = 0u; (res == yami_ok) && (i != num_of_entries); ++i)
            {
                /* read name of field */
                
                res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                if (res == yami_ok)
                {
                    if (len > 0)
                    {
                        if ((size_t)len < MAX_HEADER_TYPE_LENGTH)
                        {
                            res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                name, (size_t)len);
                        }
                        else
                        {
                            res = yami_not_enough_space;
                        }
                    }
                    else
                    {
                        res = yami_unexpected_value;
                    }
                }
                
                /* read type of field */
                
                if (res == yami_ok)
                {
                    res = yami_get_type(buffer, buffer_size, buffer_position, &type);
                }
                
                /* read field values according to their names and types */
                
                if (res == yami_ok)
                {
                    if ((yami_strncmp(name, type_tag, type_tag_len) == 0) &&
                        (type == yami_string))
                    {
                        /* this is "type" field */

                        res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                        if (res == yami_ok)
                        {
                            if (len > 0)
                            {
                                if ((size_t)len < header_type_size)
                                {
                                    res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                        out_header_type, (size_t)len);
                                }
                                else
                                {
                                    res = yami_not_enough_space;
                                }
                            }
                            else
                            {
                                res = yami_unexpected_value;
                            }
                        }
                    }
                    else if ((yami_strncmp(name, object_name_tag, object_name_tag_len) == 0) &&
                        (type == yami_string))
                    {
                        /* this is "object_name" field */

                        res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                        if (res == yami_ok)
                        {
                            if (len > 0)
                            {
                                if ((size_t)len < object_name_size)
                                {
                                    res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                        out_object_name, (size_t)len);
                                }
                                else
                                {
                                    res = yami_not_enough_space;
                                }
                            }
                            else
                            {
                                res = yami_unexpected_value;
                            }
                        }
                    }
                    else if ((yami_strncmp(name, message_name_tag, message_name_tag_len) == 0) &&
                        (type == yami_string))
                    {
                        /* this is "message_name" field */

                        res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                        if (res == yami_ok)
                        {
                            if (len > 0)
                            {
                                if ((size_t)len < message_name_size)
                                {
                                    res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                        out_message_name, (size_t)len);
                                }
                                else
                                {
                                    res = yami_not_enough_space;
                                }
                            }
                            else
                            {
                                res = yami_unexpected_value;
                            }
                        }
                    }
                    else if ((yami_strncmp(name, reason_tag, reason_tag_len) == 0) &&
                        (type == yami_string))
                    {
                        /* this is "reason" field */

                        res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                        if (res == yami_ok)
                        {
                            if (len > 0)
                            {
                                if ((size_t)len < reason_size)
                                {
                                    res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                        out_reason, (size_t)len);
                                }
                                else
                                {
                                    res = yami_not_enough_space;
                                }
                            }
                            else
                            {
                                res = yami_unexpected_value;
                            }
                        }
                    }
                    else if ((yami_strncmp(name, message_id_tag, message_id_tag_len) == 0) &&
                        (type == yami_long_long))
                    {
                        /* this is "message_id" field */

                        res = yami_get_long_long(buffer, buffer_size, buffer_position, out_message_id);
                    }
                    else
                    {
                        /* unknown, but legal (optional) field */
                        
                        res = yami_skip_field(buffer, buffer_size, buffer_position, type);
                    }
                }
            }
        }
        else
        {
            res = yami_unexpected_value;
        }
    }
    
    return res;
}

/* helper, does not check for buffer size */
static void yami_put_int32_t(uint8_t buffer[], size_t * const buffer_position,
    const int32_t value)
{
    uint8_t byte0;
    uint8_t byte1;
    uint8_t byte2;
    uint8_t byte3;
    
    uint32_t uvalue;
    if (value >= 0)
    {
        uvalue = (uint32_t)value;
    }
    else
    {
        const int32_t tmp = -1 * (value + 1);
        uvalue = (uint32_t)tmp ^ 0xffffffffu;
    }
    
    byte0 = (uint8_t)(uvalue & 0xffu);
    byte1 = (uint8_t)((uvalue & 0xff00u) >> 8u);
    byte2 = (uint8_t)((uvalue & 0xff0000u) >> 16u);
    byte3 = (uint8_t)((uvalue & 0xff000000u) >> 24u);
    
    buffer[*buffer_position] = byte0;
    buffer[*buffer_position + 1u] = byte1;
    buffer[*buffer_position + 2u] = byte2;
    buffer[*buffer_position + 3u] = byte3;

    *buffer_position += 4u;
}

/* helper, does not check for buffer size */
static void yami_get_int32_t(const uint8_t buffer[], size_t * const buffer_position,
    int32_t * const value)
{
    const uint8_t byte0 = buffer[*buffer_position];
    const uint8_t byte1 = buffer[*buffer_position + 1u];
    const uint8_t byte2 = buffer[*buffer_position + 2u];
    const uint8_t byte3 = buffer[*buffer_position + 3u];

    const uint32_t uvalue =
        ((uint32_t)byte3 << 24) |
        ((uint32_t)byte2 << 16) |
        ((uint32_t)byte1 << 8) |
        (uint32_t)byte0;
    
    if ((uvalue & 0x80000000u) == 0u)
    {
        /* positive number */
        
        *value = (int32_t)uvalue;
    }
    else
    {
        /* negative number */
        
        const uint32_t temp = uvalue ^ 0xffffffffu;
        *value = (int32_t)temp - 1;
    }
    
    *buffer_position += 4u;
}

static enum yami_result yami_fill_frame_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const uint32_t message_id,
    const size_t message_header_size, const size_t frame_payload_size)
{
    /* frame header contains 4 int fields: */
    /* - message id                        */
    /* - frame id, always -1               */
    /* - message header size               */
    /* - frame payload size                */

    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < FRAME_HEADER_SIZE)
    {
        res = yami_not_enough_space;
    }
    else
    {
        yami_put_int32_t(buffer, buffer_position, (int32_t)message_id);
        yami_put_int32_t(buffer, buffer_position, -1);
        yami_put_int32_t(buffer, buffer_position, (int32_t)message_header_size);
        yami_put_int32_t(buffer, buffer_position, (int32_t)frame_payload_size);
        
        res = yami_ok;
    }
    
    return res;
}

static enum yami_result yami_parse_frame_header(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    uint32_t * const message_id,
    size_t * const message_header_size, size_t * const frame_payload_size)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < FRAME_HEADER_SIZE)
    {
        res = yami_not_enough_space;
    }
    else
    {
        int32_t message_id_tmp;
        int32_t frame_id;
        int32_t message_header_size_tmp;
        int32_t frame_payload_size_tmp;
        
        yami_get_int32_t(buffer, buffer_position, &message_id_tmp);
        yami_get_int32_t(buffer, buffer_position, &frame_id);
        yami_get_int32_t(buffer, buffer_position, &message_header_size_tmp);
        yami_get_int32_t(buffer, buffer_position, &frame_payload_size_tmp);
        
        if ((frame_id == -1) && (message_id_tmp >= 0) &&
            (message_header_size_tmp > 0) && (frame_payload_size_tmp > 0))
        {
            *message_id = (uint32_t)message_id_tmp;
            *message_header_size = (uint32_t)message_header_size_tmp;
            *frame_payload_size = (uint32_t)frame_payload_size_tmp;
            
            res = yami_ok;
        }
        else
        {
            /* this version of YAMI4 does not support multi-frame messages */
            /* and negative payload values are strange anyway              */
            
            res = yami_unexpected_value;
        }
    }
    
    return res;
}

static void yami_channel_init(struct yami_channel * const self)
{
    self->input_data_read_so_far_ = 0u;
    self->input_message_complete_ = 0; /* false */
}

static yami_result yami_send_packet(struct yami_channel * const self,
    EthernetUDP * transport, IPAddress ip, uint16_t port, size_t data_size)
{
    enum yami_result res;
    int32_t sent;

    transport->beginPacket(ip, port);
    
    transport->write(self->output_buffer_, data_size);
    
    sent = transport->endPacket();
    if (sent != 0)
    {
        res = yami_ok;
    }
    else
    {
        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_channel_post_message(struct yami_channel * const self,
    EthernetUDP * transport, IPAddress ip, uint16_t port,
    const char object_name[], const char message_name[],
    const int64_t message_id, const uint32_t core_message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    /* serialize proper message header */
    
    size_t buffer_position = FRAME_HEADER_SIZE;
    
    res = yami_fill_message_header(
        self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
        &buffer_position,
        object_name, message_name, message_id);
    if (res == yami_ok)
    {
        const size_t header_size = buffer_position - FRAME_HEADER_SIZE;
        const size_t frame_payload_size = yami_round_up_to_4(header_size + body_size);
        
        /* copy the body buffer */
        
        yami_memcpy(self->output_buffer_, buffer_position, body, 0u, body_size);
        
        /* and fill the frame header */
        
        buffer_position = 0u;
        
        res = yami_fill_frame_header(self->output_buffer_, FRAME_HEADER_SIZE,
            &buffer_position, core_message_id, header_size, frame_payload_size);
        
        if (res == yami_ok)
        {
            size_t data_size = FRAME_HEADER_SIZE + frame_payload_size;
            
            res = yami_send_packet(self, transport, ip, port, data_size);
        }
    }
    
    return res;
}

static enum yami_result yami_channel_post_reply(struct yami_channel * const self,
    EthernetUDP * transport, IPAddress ip, uint16_t port,
    const int64_t message_id, const uint32_t core_message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    /* serialize proper message header */
    
    size_t buffer_position = FRAME_HEADER_SIZE;
    
    res = yami_fill_reply_header(
        self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
        &buffer_position, message_id);
    if (res == yami_ok)
    {
        const size_t header_size = buffer_position - FRAME_HEADER_SIZE;
        const size_t frame_payload_size = yami_round_up_to_4(header_size + body_size);
        
        /* copy the body buffer */
        
        yami_memcpy(self->output_buffer_, buffer_position, body, 0u, body_size);
        
        /* and fill the frame header */
        
        buffer_position = 0u;
        
        res = yami_fill_frame_header(self->output_buffer_, FRAME_HEADER_SIZE,
            &buffer_position, core_message_id, header_size, frame_payload_size);
        
        if (res == yami_ok)
        {
            size_t data_size = FRAME_HEADER_SIZE + frame_payload_size;
            
            res = yami_send_packet(self, transport, ip, port, data_size);
        }
    }
    
    return res;
}

static enum yami_result yami_channel_post_reject(struct yami_channel * const self,
    EthernetUDP * transport, IPAddress ip, uint16_t port,
    const int64_t message_id, const uint32_t core_message_id,
    const char reason[])
{
    enum yami_result res;
    
    /* serialize proper message header */
    
    size_t buffer_position = FRAME_HEADER_SIZE;
    
    res = yami_fill_exception_header(
        self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
        &buffer_position, reason, message_id);
    if (res == yami_ok)
    {
        /* the body buffer is empty - that is, it contains 0 elements */
        
        const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
        const size_t empty_body_size = sizeof(empty_body);
        
        const size_t header_size = buffer_position - FRAME_HEADER_SIZE;
        const size_t frame_payload_size = header_size + empty_body_size;
        
        yami_memcpy(self->output_buffer_, buffer_position, empty_body, 0u, empty_body_size);
        
        /* and fill the frame header */
        
        buffer_position = 0u;
        
        res = yami_fill_frame_header(self->output_buffer_, FRAME_HEADER_SIZE,
            &buffer_position, core_message_id, header_size, frame_payload_size);
        
        if (res == yami_ok)
        {
            size_t data_size = FRAME_HEADER_SIZE + frame_payload_size;
            
            res = yami_send_packet(self, transport, ip, port, data_size);
        }
    }
    
    return res;
}

static enum yami_result yami_channel_check_header(
    struct yami_channel * const self,
    size_t * const out_expected_packet_size)
{
    enum yami_result res = yami_ok;

    /* first read message header, then payload */
    if (self->input_data_read_so_far_ < FRAME_HEADER_SIZE)
    {
        /* message header was not yet read */
        
        *out_expected_packet_size = FRAME_HEADER_SIZE;
    }
    else
    {
        /* message header was already read, expect more for frame payload */
        
        size_t position = 0u;
        size_t * const buffer_position = &position;
        
        res = yami_parse_frame_header(
            self->input_buffer_, FRAME_HEADER_SIZE, buffer_position,
            &self->input_message_id_, &self->input_message_header_size_,
            &self->input_message_payload_size_);
        
        if (res == yami_ok)
        {
            /* prepare to read the rest of the frame */
            
            *out_expected_packet_size =
                FRAME_HEADER_SIZE + self->input_message_payload_size_;
        }
    }

    return res;
}

static enum yami_result yami_channel_check_body(
    struct yami_channel * const self,
    const size_t expected_packet_size)
{
    enum yami_result res = yami_unexpected_value;

    if (self->input_data_read_so_far_ == expected_packet_size)
    {
        if (expected_packet_size > FRAME_HEADER_SIZE)
        {
            /* complete message has been read (frame with payload) */
            /* mark the input buffer as ready for consumption      */
                        
            self->input_message_complete_ = 1; /* true */

            res = yami_ok;
        }
    }

    return res;
}

static void yami_channel_clean_input_buffer(struct yami_channel * const self)
{
    self->input_data_read_so_far_ = 0u;
    self->input_message_complete_ = 0; /* false */
}

static enum yami_result yami_channel_check_udp_packet(
    struct yami_channel * const self)
{
    size_t expected_packet_size;

    enum yami_result res = yami_channel_check_header(
        self, &expected_packet_size);
    if (res == yami_ok)
    {
        res = yami_channel_check_body(self, expected_packet_size);
        if (res != yami_ok)
        {
            /* if the packet is not complete, silently drop it */

            yami_channel_clean_input_buffer(self);
        }
    }

    return res;
}

static enum yami_result yami_channel_receive_packet(
    struct yami_channel * const self,
    EthernetUDP * transport)
{
    enum yami_result res;
    
    int32_t available = transport->available();
    
    if ((size_t)available <= MAX_INCOMING_MESSAGE_PAYLOAD)
    {
        int32_t readn = transport->read(self->input_buffer_, available);
        
        if (readn == available)
        {
            self->incomingIp_ = transport->remoteIP();
            self->incomingPort_ = transport->remotePort();
            
            /* the UDP packet is read in its entirety, without */
            /* separate stages for header and body             */
            /* - verify both parts now                         */
            
            self->input_data_read_so_far_ = readn;
            
            res = yami_channel_check_udp_packet(self);
        }
        else
        {
            res = yami_io_error;
        }
    }
    else
    {
        res = yami_unexpected_value;
    }
    
    return res;
}

static void yami_init(struct yami_agent * const self, EthernetUDP * udp)
{
    yami_channel_init(&self->channel_);
    
    self->last_core_message_id_ = 0u;
    self->last_message_id_ = 0LL;
    
    self->message_callback_ = NULL;

    self->transport_ = udp;
}

static void yami_set_message_callback(struct yami_agent * const self,
    const yami_message_function message_callback)
{
    self->message_callback_ = message_callback;
}

static enum yami_result yami_post_message(struct yami_agent * const self,
    IPAddress ip, uint16_t port,
    const char object_name[], const char message_name[],
    const uint8_t body[], const size_t body_size,
    int64_t * const out_message_id)
{
    enum yami_result res;
    
    ++self->last_core_message_id_;
    
    ++self->last_message_id_;
    *out_message_id = self->last_message_id_;
    
    res = yami_channel_post_message(&self->channel_,
        self->transport_, ip, port,
        object_name, message_name,
        self->last_message_id_, self->last_core_message_id_,
        body, body_size);
   
    return res;
}

static enum yami_result yami_post_reply(struct yami_agent * const self,
    IPAddress ip, uint16_t port,
    const int64_t message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    ++self->last_core_message_id_;
    
    res = yami_channel_post_reply(&self->channel_,
        self->transport_, ip, port,
        message_id, self->last_core_message_id_,
        body, body_size);
   
    return res;
}

static enum yami_result yami_post_reject(struct yami_agent * const self,
    IPAddress ip, uint16_t port,
    const int64_t message_id, const char reason[])
{
    enum yami_result res;
    
    ++self->last_core_message_id_;
    
    res = yami_channel_post_reject(&self->channel_,
        self->transport_, ip, port,
        message_id, self->last_core_message_id_,
        reason);
   
    return res;
}

static enum yami_result yami_dispatch_incoming_message(
    const struct yami_agent * const self,
    struct yami_channel * const channel)
{
    enum yami_result res = yami_ok;
    
    const size_t body_size = channel->input_message_payload_size_ -
        channel->input_message_header_size_;

    if (self->message_callback_ != NULL)
    {
        /* high-level callback -> parse the message header */
            
        int64_t message_id;
        char header_type[MAX_HEADER_TYPE_LENGTH];
        char object_name[MAX_OBJECT_NAME_LENGTH];
        char message_name[MAX_MESSAGE_NAME_LENGTH];
        char reason[MAX_EXCEPTION_REASON_LENGTH];
            
        const char message_tag[] = "message";
        const size_t message_tag_len = sizeof(message_tag);
        const char reply_tag[] = "reply";
        const size_t reply_tag_len = sizeof(reply_tag);
        const char exception_tag[] = "exception";
        const size_t exception_tag_len = sizeof(exception_tag);

        const char empty_message_name[1] = "";
        const char empty_object_name[1] = "";
        const char empty_exception_msg[1] = "";
        
        const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
        const size_t empty_body_size = sizeof(empty_body);

        size_t rpos = FRAME_HEADER_SIZE;
            
        res = yami_parse_message_header(
            channel->input_buffer_, FRAME_HEADER_SIZE + channel->input_message_header_size_, &rpos,
            header_type, MAX_HEADER_TYPE_LENGTH,
            object_name, MAX_OBJECT_NAME_LENGTH,
            message_name, MAX_MESSAGE_NAME_LENGTH,
            reason, MAX_EXCEPTION_REASON_LENGTH,
            &message_id);
            
        if (res == yami_ok)
        {
            if (yami_strncmp(header_type, message_tag, message_tag_len) == 0)
            {
                /* this is a regular incoming message */
                    
                self->message_callback_(
                    channel->incomingIp_, channel->incomingPort_,
                    yami_incoming_message,
                    object_name, message_name, empty_exception_msg,
                    message_id,
                    &channel->input_buffer_[FRAME_HEADER_SIZE + channel->input_message_header_size_],
                    body_size);
            }
            else if (yami_strncmp(header_type, reply_tag, reply_tag_len) == 0)
            {
                /* this is a reply message */
                    
                self->message_callback_(
                    channel->incomingIp_, channel->incomingPort_,
                    yami_message_reply,
                    empty_object_name, empty_message_name, empty_exception_msg,
                    message_id,
                    &channel->input_buffer_[FRAME_HEADER_SIZE + channel->input_message_header_size_],
                    body_size);
            }
            else if (yami_strncmp(header_type, exception_tag, exception_tag_len) == 0)
            {
                /* this is a rejection (exception) message */
                    
                self->message_callback_(
                    channel->incomingIp_, channel->incomingPort_,
                    yami_message_rejected,
                    empty_object_name, empty_message_name, reason,
                    message_id,
                    empty_body, empty_body_size);
            }
            else
            {
                res = yami_unexpected_value;
            }
        }
    }

    if (res == yami_ok)
    {
        /* allow new messages to arrive on this channel */

        yami_channel_clean_input_buffer(channel);
    }
    
    return res;
}

static enum yami_result yami_receive(struct yami_agent * const self)
{
    enum yami_result res;

    int32_t packet_size = self->transport_->parsePacket();
    if (packet_size != 0)
    {
        res = yami_channel_receive_packet(&self->channel_, self->transport_);
        
        if ((res == yami_ok) &&
            (self->channel_.input_message_complete_ != 0))
        {
            /* dispatch this incoming message */
            
            res = yami_dispatch_incoming_message(self, &self->channel_);
        }
    }
    else
    {
        res = yami_ok;
    }
    
    return res;
}

YAMI4Agent::YAMI4Agent()
{
}

void YAMI4Agent::init(EthernetUDP & udp, yami_message_function message_callback)
{
    yami_init(&agent_, &udp);

    yami_set_message_callback(&agent_, message_callback);
}

void YAMI4Agent::init(EthernetUDP & udp)
{
    yami_init(&agent_, &udp);
}

enum yami_result YAMI4Agent::sendMessage(
    IPAddress ip, uint16_t port,
    const char object_name[], const char message_name[],
    const uint8_t body[], size_t body_size,
    int64_t & out_message_id)
{
    return yami_post_message(&agent_, ip, port, object_name, message_name,
        body, body_size, &out_message_id);
}

enum yami_result YAMI4Agent::sendReply(
    IPAddress ip, uint16_t port,
    int64_t message_id,
    const uint8_t body[], size_t body_size)
{
    return yami_post_reply(&agent_, ip, port, message_id,
        body, body_size);
}

enum yami_result YAMI4Agent::sendReject(
    IPAddress ip, uint16_t port,
    int64_t message_id, const char reason[])
{
    return yami_post_reject(&agent_, ip, port, message_id,
        reason);
}

enum yami_result YAMI4Agent::receive()
{
    return yami_receive(&agent_);
}
