// Copyright Inspirel Sp. z o.o. 2014.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#ifndef YAMI4ARDUINO_H
#define YAMI4ARDUINO_H

#include <stddef.h>
#include <stdint.h>

/* max size of output message payload */
#define MAX_OUTGOING_MESSAGE_PAYLOAD 250u

/* max size of input message payload */
#define MAX_INCOMING_MESSAGE_PAYLOAD 250u

/* Note: the following value:
 *
 *     MAX_OUTGOING_MESSAGE_PAYLOAD + MAX_INCOMING_MESSAGE_PAYLOAD
 *
 * is the most significant contributor
 * to the static memory consumption of the YAMI4 agent.
 */

/// General type for reporting success and error states.
enum yami_result
{
    yami_ok,                ///< Operation completed successfully.
    yami_no_such_name,      ///< The given name was not found.
    yami_not_enough_space,  ///< There is not enough space in the buffer.
    yami_unexpected_value,  ///< The given value was not recognized.
    yami_bad_protocol,      ///< The connection protocol is incorrect.
    yami_io_error,          ///< Unable to perform the I/O operation.
    yami_timed_out,         ///< The requested operation timed out.
    yami_channel_closed,    ///< The operation was not possible due to EOF.
    yami_bad_state          ///< The given object is in the wrong state.
};

/// Type of parameter entry.
enum yami_parameter_type
{
    yami_boolean,                ///< bool
    yami_integer,                ///< int
    yami_long_long,              ///< long long
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    yami_double_float,           ///< double
#endif // YAMI4_WITH_DOUBLE_FLOAT
    yami_string,                 ///< string
    yami_binary,                 ///< Binary block
    yami_boolean_array,          ///< Array of bool
    yami_integer_array,          ///< Array of int
    yami_long_long_array,        ///< Array of long long
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    yami_double_float_array,     ///< Array of double
#endif // YAMI4_WITH_DOUBLE_FLOAT
    yami_string_array,           ///< Array of strings
    yami_binary_array,           ///< Array of binary blocks
    yami_nested_parameters,      ///< Nested parameters object
    yami_nested_parameters_array ///< Array of nested parameters object
};

/// Type of message for high-level message delivery.
enum yami_message_type
{
    yami_incoming_message,  ///< new incoming message
    yami_message_reply,     ///< reply from remote receiver
    yami_message_rejected   ///< outgoing message was rejected
};

/// Type of function callback for high-level message delivery.
///
/// @param target endpoint name of the message sender
/// @param message_type type of message notification
/// @param object_name name of logical message destination
///                    (valid only for yami_incoming_message)
/// @param message_name name of message
///                     (valid only for yami_incoming_message)
/// @param exception_msg description of problem at the receiver side
///                      (valid only for yami_incoming_message)
/// @param message_id high-level message identifier
/// @param body body buffer
///             (valid only for yami_incoming_message and for yami_message_reply)
/// @param body_size size of the body buffer
///                  (valid only for yami_incoming_message and for yami_message_reply)
///
typedef void (*yami_message_function)(
    IPAddress remoteIp, uint16_t remotePort,
    enum yami_message_type message_type,
    const char object_name[], const char message_name[], const char exception_msg[],
    int64_t message_id,
    const uint8_t body[], size_t body_size);

/// \brief Places type indicator value in the serialization buffer.
///
/// Places type indicator value in the serialization buffer
/// at the specified index and advances the index to the position
/// immediately after the written value.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param type type code to be written
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_type(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type type);

/// \brief Reads type code from the serialization buffer.
///
/// Reads type code from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param type (out) type code to be read
///
/// @return <code>yami_ok</code> or <code>yami_unexpected_value</code>
///         or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_type(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type * type);

/// \brief Places integer value in the serialization buffer.
///
/// Places integer value in the serialization buffer at the specified index and
/// advances the index to the position immediately after the written value.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value integer value to be written
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_integer(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value);

/// \brief Reads integer value from the serialization buffer.
///
/// Reads integer value from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) integer value to be read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_integer(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t * value);

/// \brief Places long long value in the serialization buffer.
///
/// Places long long value in the serialization buffer at the specified index and
/// advances the index to the position immediately after the written value.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value long long value to be written
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_long_long(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t value);

/// \brief Reads long long value from the serialization buffer.
///
/// Reads long long value from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) long long value to be read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_long_long(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t * value);

#ifdef YAMI4_WITH_DOUBLE_FLOAT

/// \brief Places double float value in the serialization buffer.
///
/// Places double float value in the serialization buffer at the specified index and
/// advances the index to the position immediately after the written value.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value double float value to be written
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_double_float(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value);

/// \brief Reads double float value from the serialization buffer.
///
/// Reads double float value from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) double float value to be read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_double_float(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double * value);

#endif // YAMI4_WITH_DOUBLE_FLOAT

/// \brief Places string value in the serialization buffer.
///
/// Places string value in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: It is assumed that the string is nul-terminated and the function
///       will stop copying characters at the first nul character.
///       The function stores the length prefix before the string itself;
///       the length does not take into account the nul character.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value string value to be written, nul-terminated
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_cstring(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const char value[]);

/// \brief Reads raw string value from the serialization buffer.
///
/// Reads raw string value from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The string is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output, together with the nul character
///       that is appended after the extracted value.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the string
/// @param value_length length of string to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_cstring(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    char value[], size_t value_length);

/// \brief Places binary value in the serialization buffer.
///
/// Places binary value in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: The function stores the length prefix before the binary itself.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value binary value to be written
/// @param value_length length of the binary value
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_binary(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const uint8_t value[], size_t value_length);

/// \brief Reads raw binary value from the serialization buffer.
///
/// Reads raw binary value from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The binary is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the binary value
/// @param value_length length of binary to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_binary(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    uint8_t value[], size_t value_length);

/// \brief Places boolean array in the serialization buffer.
///
/// Places boolean array in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: The function stores the length prefix before the array itself.
///       The boolean array is stored in the packed form, properly padded.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value boolean array to be written,
///              each array field contains a single boolean value
/// @param value_length length of the array
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_boolean_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int32_t value[], size_t value_length);

/// \brief Reads raw boolean array from the serialization buffer.
///
/// Reads raw boolean array from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The array is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output.
///       The array is read in an unpacked form.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the array value
/// @param value_length length of array to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_boolean_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value[], size_t value_length);

/// \brief Places integer array in the serialization buffer.
///
/// Places integer array in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: The function stores the length prefix before the array itself.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value integer array to be written
/// @param value_length length of the array
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_integer_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int32_t value[], size_t value_length);

/// \brief Reads raw integer array from the serialization buffer.
///
/// Reads raw integer array from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The array is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the array value
/// @param value_length length of array to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_integer_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value[], size_t value_length);

/// \brief Places long long array in the serialization buffer.
///
/// Places long long array in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: The function stores the length prefix before the array itself.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value long long array to be written
/// @param value_length length of the array
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_long_long_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int64_t value[], size_t value_length);

/// \brief Reads raw long long array from the serialization buffer.
///
/// Reads raw long long array from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The array is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the array value
/// @param value_length length of array to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_long_long_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t value[], size_t value_length);

#ifdef YAMI4_WITH_DOUBLE_FLOAT

/// \brief Places double float array in the serialization buffer.
///
/// Places double float array in the serialization buffer at the specified index and
/// advances the index to the position after the written value.
/// Note: The function stores the length prefix before the array itself.
///
/// @param buffer (in/out) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the new value should be written
/// @param value double float array to be written
/// @param value_length length of the array
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_put_double_float_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const double value[], size_t value_length);

/// \brief Reads raw double float array from the serialization buffer.
///
/// Reads raw double float array from the serialization buffer from the specified index and
/// advances the index to the position immediately after the value that was read.
/// Note: The array is read without the length prefix, so it is assumed that the caller
///       has already read the length prefix and has allocated the buffer
///       that is big enough to accommodate the output.
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
///                        where the value should be read from
/// @param value (out) buffer to store the array value
/// @param value_length length of array to read
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_get_raw_double_float_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value[], size_t value_length);

#endif // YAMI4_WITH_DOUBLE_FLOAT

/// \brief Skips the given field in serialization buffer.
///
/// Skips the field of the given type in the serialization buffer.
/// This function can be particularly useful when processing fields
/// that can have varying length (strings, binaries and arrays).
///
/// @param buffer (in) serialization buffer
/// @param buffer_size size of the serialization buffer
/// @param buffer_position (in/out) position (in bytes) in the buffer array
/// @param type type of the field to be skipped
///
/// @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
///
enum yami_result yami_skip_field(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type type);

// for internal definition:

struct yami_channel
{
    size_t output_data_size_;
    
    IPAddress incomingIp_;
    uint16_t incomingPort_;
    size_t input_data_read_so_far_;
    uint32_t input_message_id_;
    size_t input_message_header_size_;
    size_t input_message_payload_size_;
    int32_t input_message_complete_;

    /* note: sizes of the arrays below can interact */
    /* with required padding at the end of struct   */
    
    uint8_t output_buffer_[MAX_OUTGOING_MESSAGE_PAYLOAD];
    uint8_t input_buffer_[MAX_INCOMING_MESSAGE_PAYLOAD];
};

struct yami_agent
{
    uint32_t last_core_message_id_;
    int64_t last_message_id_;
    
    yami_message_function message_callback_;
    
    struct yami_channel channel_;

    EthernetUDP * transport_;
};

/// \brief Message broker.
///
/// The message broker that encapsulates physical channel management,
/// incoming and outgoing message buffers, and
/// related resource management activities.
/// 
/// The agent objects can be created and destroyed without constraints
/// on the stack, on the free store or as static objects.
/// 
class YAMI4Agent
{
public:

    /// \brief Agent constructor.
    ///
    /// Agent constructor, creates an uninitialized object, which should
    /// be properly initialized by a separate call to init().
    ///
    YAMI4Agent();

    /// \brief Initializes the agent object.
    ///
    /// Initializes the agent object with the given transport and callback.
    ///
    /// @param udp transport object, already prepared and configured
    /// @param message_callback function that will receive messages
    ///
    void init(EthernetUDP & udp, yami_message_function message_callback);

    /// \brief Initializes the agent object.
    ///
    /// Initializes the agent object with the given transport and no callback.
    /// Such an object can be only used for messaging sending.
    ///
    /// @param udp transport object, already prepared and configured
    ///
    void init(EthernetUDP & udp);

    /// \brief Sends the high-level message.
    ///
    /// Sends the high-level message.
    ///
    /// @param ip remote target address
    /// @param port remote target port
    /// @param object_name name of logical message destination
    /// @param message_name name of message
    /// @param body already serialized body buffer
    /// @param body_size size of the body buffer
    /// @param out_message_id high-level message identifier, allocated by this agent
    /// @return <code>yami_ok</code> if the message was posted,
    ///         <code>yami_no_such_name</code> if the given target does not
    ///                                        reference any open channel or
    ///         <code>yami_not_enough_space</code> if the message is bigger
    ///                                            than the buffer space
    ///
    enum yami_result sendMessage(IPAddress ip, uint16_t port,
        const char object_name[], const char message_name[],
        const uint8_t body[], size_t body_size,
        int64_t & out_message_id);

    /// \brief Sends the high-level message reply.
    ///
    /// Sends the high-level message reply.
    ///
    /// @param ip remote target address
    /// @param port remote target port
    /// @param message_id high-level message identifier, allocated by remote sender
    /// @param body already serialized body buffer
    /// @param body_size size of the body buffer
    /// @return <code>yami_ok</code> if the message was posted,
    ///         <code>yami_no_such_name</code> if the given target does not
    ///                                        reference any open channel or
    ///         <code>yami_not_enough_space</code> if the message is bigger
    ///                                            than the buffer space
    ///
    enum yami_result sendReply(IPAddress ip, uint16_t port,
        int64_t message_id,
        const uint8_t body[], size_t body_size);

    /// \brief Sends the high-level message rejection.
    ///
    /// Sends the high-level message rejection.
    ///
    /// @param ip remote target address
    /// @param port remote target port
    /// @param message_id high-level message identifier, allocated by remote sender
    /// @param reason exception message, intended as description of
    ///               why the original message is rejected
    /// @return <code>yami_ok</code> if the message was posted,
    ///         <code>yami_no_such_name</code> if the given target does not
    ///                                        reference any open channel or
    ///         <code>yami_not_enough_space</code> if the message is bigger
    ///                                            than the buffer space
    ///
    enum yami_result sendReject(IPAddress ip, uint16_t port,
        int64_t message_id, const char reason[]);

    /// \brief Parses and dispatches incoming packet if any is available.
    ///
    /// @return <code>yami_ok</code> if the work is performed
    ///                              or there was nothing to do
    ///         <code>yami_unexpected_value</code> if the incoming package cannot be parsed
    ///         <code>yami_io_error</code> in case of system-related error
    ///
    enum yami_result receive();

private:

    struct yami_agent agent_;

};

#endif // YAMI4ARDUINO_H
