//
// This calculator example receives two numbers and
// sends back results of four basic calculations.
// It can work with any "calculator" client example from the YAMI4 package.
//

// Note: these headers have to be included in this order:
#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h>
#include <YAMI4.h>

// set up your MAC and IP address:
byte mac[] = {  
  0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };
IPAddress ip(192, 168, 1, 177);

// local port to listen on:
unsigned int localPort = 12345;

EthernetUDP udp;
YAMI4Agent agent;

// this function will receive all YAMI4 messages:
void call(IPAddress remoteIp, uint16_t remotePort,
    enum yami_message_type message_type,
    const char object_name[], const char message_name[], const char exception_msg[],
    int64_t message_id,
    const uint8_t body[], size_t body_size)
{
  // ignore the header, but deserialize the message body

  enum yami_result res;
  
  int32_t num_of_entries;
  int32_t i;
  int32_t len;
  char field_name[2]; // expect only "a" and "b"
  enum yami_parameter_type type;

  int32_t a; // request parameter "a"
  int32_t b; // request parameter "b"
  const size_t reply_buffer_size = 100u;
  uint8_t reply_body[reply_buffer_size]; // serialized reply body
  size_t reply_body_size; // actual size of serialized reply
    
  // buffer position is updated with each reading operation
  size_t buffer_position = 0u;
  
  // for simplicity this server does not allow unknown fields,
  // so the only expected message format is a regular message
  // with "a" and "b" integer fields
  // - any other message will be silently ignored
  
  res = yami_get_integer(body, body_size, &buffer_position, &num_of_entries);
  
  if ((res != yami_ok) || (num_of_entries != 2))
  {
    return;
  }
  
  for (i = 0; i != num_of_entries; ++i)
  {
    // read length of field name
    
    res = yami_get_integer(body, body_size, &buffer_position, &len);
        
    if ((res != yami_ok) || (len != 1))
    {
      return;
    }
    
    // read the field name (expect "a" or "b")
      
    res = yami_get_raw_cstring(body, body_size, &buffer_position,
        field_name, (size_t)len);
    if (res != yami_ok)
    {
      return;
    }
    
    // read type (expect integer for both fields)
    
    res = yami_get_type(body, body_size, &buffer_position, &type);
    
    if ((res != yami_ok) || (type != yami_integer))
    {
      return;
    }
    
    if (strcmp(field_name, "a") == 0)
    {
      // read "a" integer value
      
      res = yami_get_integer(body, body_size, &buffer_position, &a);
    }
    else
    {
      // read "b" integer value
      
      res = yami_get_integer(body, body_size, &buffer_position, &b);
    }

    if (res != yami_ok)
    {
      return;
    }
  }
  
  // at this point the whole message was parsed:
  // compute the results and prepare reply body in appropriate format
  
  // note: if b == 0, the ratio field will not be present in the reply
  
  if (b != 0)
  {
    num_of_entries = 4;
  }
  else
  {
    num_of_entries = 3;
  }
  
  // reply_body_size will be updated with each serialized element
  // and at the end will indicate the actual size of the whole message
  reply_body_size = 0u;

  // number of fields
  res = yami_put_integer(reply_body, reply_buffer_size, &reply_body_size, num_of_entries);
  if (res != yami_ok)
  {
    return;
  }

  // put sum value
  
  // field name
  res = yami_put_cstring(reply_body, reply_buffer_size, &reply_body_size, "sum");
  if (res != yami_ok)
  {
    return;
  }

  // field type
  res = yami_put_type(reply_body, reply_buffer_size, &reply_body_size, yami_integer);
  if (res != yami_ok)
  {
    return;
  }
  
  // field value
  res = yami_put_integer(reply_body, reply_buffer_size, &reply_body_size, a + b);
  if (res != yami_ok)
  {
    return;
  }

  // put difference value
  
  // field name
  res = yami_put_cstring(reply_body, reply_buffer_size, &reply_body_size, "difference");
  if (res != yami_ok)
  {
    return;
  }

  // field type
  res = yami_put_type(reply_body, reply_buffer_size, &reply_body_size, yami_integer);
  if (res != yami_ok)
  {
    return;
  }
  
  // field value
  res = yami_put_integer(reply_body, reply_buffer_size, &reply_body_size, a - b);
  if (res != yami_ok)
  {
    return;
  }

  // put product value
  
  // field name
  res = yami_put_cstring(reply_body, reply_buffer_size, &reply_body_size, "product");
  if (res != yami_ok)
  {
    return;
  }

  // field type
  res = yami_put_type(reply_body, reply_buffer_size, &reply_body_size, yami_integer);
  if (res != yami_ok)
  {
    return;
  }
  
  // field value
  res = yami_put_integer(reply_body, reply_buffer_size, &reply_body_size, a * b);
  if (res != yami_ok)
  {
    return;
  }
  
  // put ratio value, if possible
  
  if (b != 0)
  {
    // field name
    res = yami_put_cstring(reply_body, reply_buffer_size, &reply_body_size, "ratio");
    if (res != yami_ok)
    {
      return;
    }

    // field type
    res = yami_put_type(reply_body, reply_buffer_size, &reply_body_size, yami_integer);
    if (res != yami_ok)
    {
      return;
    }
    
    // field value
    res = yami_put_integer(reply_body, reply_buffer_size, &reply_body_size, a / b);
    if (res != yami_ok)
    {
      return;
    }
  }
  
  // here there reply_body contains the already serialized response,
  // which can be sent back to the client
  // (note how message_id allows to match responses to requests)
  
  res = agent.sendReply(remoteIp, remotePort, message_id, reply_body, reply_body_size);
  if (res != yami_ok)
  {
    return;
  }

  Serial.print("got message with parameters ");
  Serial.print(a);
  Serial.print(" and ");
  Serial.print(b);
  Serial.println(", response has been sent back");
}

void setup()
{
  // start the Ethernet and UDP:
  Ethernet.begin(mac,ip);
  udp.begin(localPort);

  // initialize YAMI4 agent:
  agent.init(udp, &call);

  Serial.begin(9600);
}

void loop()
{
  // if there is any incoming message,
  // the call function will be executed:
  agent.receive();

  delay(100);
}
